"""Configuration API for the PHAC helpers logging configuration"""

import json
import logging.config
import os
import sys
from typing import Any, Callable, Dict, Literal

import structlog

from phac_aspc.django.settings.utils import is_running_tests

DATE_FORMAT = "%d/%b/%Y %H:%M:%S"

STRUCTLOG_PRE_PROCESSORS = (
    structlog.contextvars.merge_contextvars,
    structlog.processors.TimeStamper(fmt="iso"),
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
    structlog.processors.CallsiteParameterAdder(
        {
            structlog.processors.CallsiteParameter.FUNC_NAME,
            structlog.processors.CallsiteParameter.LINENO,
        }
    ),
    structlog.processors.UnicodeDecoder(),
)

_default_suffix = "phac_helper_"

PHAC_HELPER_CONSOLE_FORMATTER_KEY = f"{_default_suffix}console_formatter"
PHAC_HELPER_PLAIN_STRING_FORMATTER_KEY = (
    f"{_default_suffix}plaintext_formatter"
)
PHAC_HELPER_JSON_FORMATTER_KEY = f"{_default_suffix}json_formatter"
PHAC_HELPER_PRETTY_JSON_FORMATTER_KEY = (
    f"{_default_suffix}pretty_json_formatter"
)

PHAC_HELPER_CONSOLE_HANDLER_KEY = f"{_default_suffix}console_handler"

PHAC_HELPER_NO_PASS_FILTER_KEY = f"{_default_suffix}no_pass_filter"

# flag set to true when configuration function called
is_phac_helper_logging_configuration_being_used = False


def configure_uniform_std_lib_and_structlog_logging(
    lowest_level_to_log: Literal[
        "NOTSET", "DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"
    ] = "INFO",
    mute_console_handler: bool = is_running_tests(),
    console_handler_formatter_key: str = PHAC_HELPER_JSON_FORMATTER_KEY,
    additional_handler_configs: Dict[
        str,
        Dict[
            str,
            Any,
        ],
    ] = None,
    additional_formatter_functions: Dict[
        str,
        Callable[
            [structlog.typing.WrappedLogger, str, structlog.typing.EventDict],
            str,
        ],
    ] = None,
    additional_filter_configs: Dict[
        str,
        Dict[
            str,
            Any,
        ],
    ] = None,
):
    """Configures both structlog and the standard library logging module, enforcing
    uniform logging behaviour between the two. Log handler and formatters are shared
    between the two, and the same set of structlog processors is run on all logs.

    The baseline configuration provides a console (stdout) handler and formatters for
    basic JSON string formatting, pretty JSON string formatting, console formatting
    (with coloured text etc.), and plain text formatting. The keys for these default
    formatters are exposed as PHAC_HELPER_..._FORMATTER_KEY variables in this module.

    `additional_handler_configs` takes standard logging dict config handler definitions.

    `additional_formatter_functions` takes a dict of callables by (unique) formatter key. These
    callables are used as structlog "renderer" (end-of-chain) processors, for seralizing the
    structlog event-dict object in to a string for the handlers to emit. I recommend directly using,
    or wrapping/subclassing, existing structlog renderers here.

    `additional_filter_configs` takes standard logging dict config filter definitions.
    In Python 3.11 and above, you can instead directly provide filter functions/instances in your
    handler configs. In prior versions, you configure them separetly and reference them by key
    in your handler configs.

    Note: by default, the built in console handler is muted running tests, because it makes
    pytest's own console output harder to follow (and pytest captures and reports errors
    after all tests have finished running anyway). You can over ride this behaviour by
    explicitly passing a `mute_console_handler` value.
    """
    # Structlog configuration to effectively make it a wrapper for the sandard library `logging`
    # module, which makes our following `logging` configuration the single source of truth on
    # project logging, and means `logging.getLogger()` and `structlog.get_logger()` produce
    # consistent output (which is very nice to have when packages might be logging via either). See:
    # https://www.structlog.org/en/stable/standard-library.html#rendering-using-structlog-based-formatters-within-logging
    structlog.configure(
        processors=[
            structlog.stdlib.filter_by_level,
            *STRUCTLOG_PRE_PROCESSORS,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.stdlib.BoundLogger,
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
        # string) from the final processor (`JSONRenderer`) will be passed to
        # the method of the same name as that you've called on the bound logger.
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    class NoPassFilter:
        def filter(self, *args, **kwargs):
            return 0

    filters = {
        PHAC_HELPER_NO_PASS_FILTER_KEY: {"()": NoPassFilter},
        **(additional_filter_configs or {}),
    }

    formatter_functions = {
        PHAC_HELPER_CONSOLE_FORMATTER_KEY: structlog.dev.ConsoleRenderer(),
        PHAC_HELPER_JSON_FORMATTER_KEY: structlog.processors.JSONRenderer(),
        PHAC_HELPER_PRETTY_JSON_FORMATTER_KEY: structlog.processors.JSONRenderer(
            # JSON string with indents, sorting, and newlines inside values preserved
            serializer=lambda *args, **kwargs: json.dumps(
                *args, indent=4, sort_keys=True, **kwargs
            ).replace("\\n", "\n")
        ),
        PHAC_HELPER_PLAIN_STRING_FORMATTER_KEY: structlog.processors.LogfmtRenderer(
            sort_keys=True
        ),
        **(additional_formatter_functions or {}),
    }

    def formatter_function_to_formatter_config(formatter_function):
        return {
            "()": structlog.stdlib.ProcessorFormatter,
            "foreign_pre_chain": STRUCTLOG_PRE_PROCESSORS,
            "datefmt": DATE_FORMAT,
            "processor": formatter_function,
        }

    formatters = {
        formatter_key: formatter_function_to_formatter_config(
            formatter_function
        )
        for (formatter_key, formatter_function) in formatter_functions.items()
    }

    handlers = {
        PHAC_HELPER_CONSOLE_HANDLER_KEY: {
            "class": "logging.StreamHandler",
            "level": lowest_level_to_log,
            "stream": (
                # pylint: disable=consider-using-with
                open(os.devnull, "w", encoding="UTF-8")
                if mute_console_handler
                else sys.stdout
            ),
            "formatter": console_handler_formatter_key,
        },
        **(additional_handler_configs or {}),
    }

    # Configure the standard library logging module
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": True,
            "filters": filters,
            "formatters": formatters,
            "handlers": handlers,
            "django.request": {
                # built in django.request logging is redundant to the django_structlog
                # request middleware, filter them out to avoid messy duplicate logs
                "filter": PHAC_HELPER_NO_PASS_FILTER_KEY
            },
            "root": {
                "level": lowest_level_to_log,
                "handlers": list(handlers.keys()),
            },
        }
    )

    # There's a separate class of warnings (e.g. warning.warn) that, by default, are handled
    # outside the logging system (even though logging has its own WARNING category). They can
    # be configured to surface as warning-level logs. No reason not to!
    logging.captureWarnings(True)

    global is_phac_helper_logging_configuration_being_used  # pylint: disable=global-statement
    is_phac_helper_logging_configuration_being_used = True



"""Provides all URLs in a single module"""

from .auth import urlpatterns as auth_patterns
from .wet import urlpatterns as wet_patterns

urlpatterns = wet_patterns + auth_patterns

__all__ = [
    "urlpatterns",
]



import inspect

from jinja2 import Environment, FileSystemLoader


class JinjaRegistry:
    def __init__(self):
        self.globals = {}
        self.filters = {}
        self.tests = {}
        self.extensions = []

    def include_default_helpers(self):
        from phac_aspc.jinja import standard_helpers

    def get_environment(self, **options):

        self.include_default_helpers()

        env = Environment(**options)
        env.globals.update(self.globals)
        env.filters.update(self.filters)
        env.tests.update(self.tests)
        for extension in self.extensions:
            env.add_extension(extension)

        return env

    def add_extension(self, extension):
        """
        Add a Jinja2 extension to the environment.
        """
        self.extensions.append(extension)

    @staticmethod
    def _get_name_and_value(name=None, value=None):
        if value is None:
            if inspect.isfunction(name) or inspect.isclass(name):
                value = name
                name = value.__name__
                return name, value
            else:
                raise ValueError(
                    "name must be provided for non-function/class objects"
                )

        return name, value

    def add_global(self, name=None, value=None):
        name, value = self._get_name_and_value(name, value)

        self.globals[name] = value

        return value

    def add_filter(self, name=None, value=None):
        name, value = self._get_name_and_value(name, value)

        self.filters[name] = value

        return value

    def add_test(self, name=None, value=None):
        name, value = self._get_name_and_value(name, value)

        self.tests[name] = value

        return value


registry = JinjaRegistry()



from glob import glob
from pathlib import Path

PHAC_ASPC_LOGGING_USE_HELPERS_CONFIG = True
# pylint: disable=wrong-import-position, wildcard-import, unused-wildcard-import
from phac_aspc.django.settings.logging import *

ALLOWED_HOSTS = ["*"]
DEBUG = True

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "testapp",
    "phac_aspc.django.helpers",
]
USE_I18N = (True,)
LANGUAGE = (
    ("fr-ca", "French"),
    ("en-ca", "English"),
)
LANGUAGE_CODE = "en-ca"
MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django_structlog.middlewares.RequestMiddleware",
]

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"
AUTH_USER_MODEL = "testapp.User"

TEST_RUNNER = "testapp.tests.pytest_runner.PytestRunner"

context_processors = [
    "django.template.context_processors.debug",
    "django.template.context_processors.request",
    "django.contrib.auth.context_processors.auth",
    "django.contrib.messages.context_processors.messages",
]


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": glob(
            str(Path(__file__).parent.joinpath("**/templates")),
            recursive=True,
        ),
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": context_processors,
        },
    },
    {
        "BACKEND": "django.template.backends.jinja2.Jinja2",
        "DIRS": glob(
            str(Path(__file__).parent.joinpath("**/jinja2")),
            recursive=True,
        ),
        "APP_DIRS": True,
        "OPTIONS": {
            "environment": "testapp.jinja2.environment",
            "context_processors": context_processors,
        },
    },
]

STATICFILES_DIRS = glob(
    str(Path(__file__).parent.joinpath("**/static")),
    recursive=True,
)


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
    }
}

SECRET_KEY = "abcdefg"


ROOT_URLCONF = "testapp.urls"
APPEND_SLASH = True



"""Recommended values related to security controls"""

from django.core.exceptions import ImproperlyConfigured

from .security_env import get_oauth_env_value, security_env_config
from .utils import (
    configure_authentication_backends,
    global_from_env,
)

# Lockout users based on their (username, IP address) combinations
AXES_LOCKOUT_PARAMETERS = [["username", "ip_address"]]

# Log unsuccessful logins
AXES_ENABLE_ACCESS_FAILURE_LOG = True

# After 3 failed login attempts, lock out the combination (IP + user).
AXES_FAILURE_LIMIT = 3

# After 30 minutes, locked out accounts are automatically unlocked.
AXES_COOLOFF_TIME = 0.5

# Reverse proxy configuration
AXES_IPWARE_PROXY_COUNT = 1  # (Behind 1 proxy)
AXES_IPWARE_META_PRECEDENCE_ORDER = [
    "HTTP_X_FORWARDED_FOR",
    "REMOTE_ADDR",
]

# Configure the identity provider if the `{PREFIX}OAUTH_PROVIDER`
# `environment variable is set.

if get_oauth_env_value("PROVIDER") == "microsoft":
    provider = get_oauth_env_value("PROVIDER")
    client_id = get_oauth_env_value("APP_CLIENT_ID")
    client_secret = get_oauth_env_value("APP_CLIENT_SECRET")
    tenant = get_oauth_env_value("MICROSOFT_TENANT")

    if client_id == "":
        raise ImproperlyConfigured("settings.OAUTH_APP_CLIENT_ID is required.")

    if client_secret == "":
        raise ImproperlyConfigured(
            "settings.OAUTH_APP_CLIENT_SECRET is required."
        )

    AUTHLIB_OAUTH_CLIENTS = {
        f"{provider}": {
            "client_id": client_id,
            "client_secret": client_secret,
            "server_metadata_url": f"https://login.microsoftonline.com/{tenant}/v2.0/.well-known/openid-configuration",  # pylint: disable=line-too-long  # noqa: E501
            "client_kwargs": {
                "scope": "openid email profile",
            },
        }
    }

#  AC-7 Automatic lockout of users after invalid login attempts
AUTHENTICATION_BACKENDS = configure_authentication_backends(
    [
        "django.contrib.auth.backends.ModelBackend",
    ]
)

#  AC-11 - Session controls set via security_env_config
global_from_env(**security_env_config)



"""From this module import * in your settings.py for a better default logging configuration"""

from phac_aspc.django.helpers.logging.configure_logging import (
    PHAC_HELPER_CONSOLE_FORMATTER_KEY,
    PHAC_HELPER_JSON_FORMATTER_KEY,
    PHAC_HELPER_PRETTY_JSON_FORMATTER_KEY,
    _default_suffix,
    configure_uniform_std_lib_and_structlog_logging,
)

from .logging_env import get_logging_env_value

if get_logging_env_value("USE_HELPERS_CONFIG"):
    # `LOGGING_CONFIG = None` drops the Django default logging config rather than merging
    # our rules with it. This is prefferable as having to consult the default rules and work out
    # what is or isn't overwritten by the merging just feels like gotcha city for future
    # maintainers. This doesn't disable built-in loggers, just let's us cleanly customize
    # the handlers and formatters that will be catching them
    LOGGING_CONFIG = None

    lowest_level_to_log = get_logging_env_value("LOWEST_LEVEL")

    additional_handler_configs = {}
    additional_filter_configs = {}

    azure_insights_connection_string = get_logging_env_value(
        "AZURE_INSIGHTS_CONNECTION_STRING"
    )
    if azure_insights_connection_string is not None:
        try:
            from opencensus.ext.azure.log_exporter import AzureLogHandler
        except (ImportError, ModuleNotFoundError) as exc:
            raise ImportError(
                "The `opencensus-ext-azure` package is required for use of PHAC helper's "
                + "Azure Insights logging. You must install this dependency your self."
            ) from exc

        additional_handler_configs[f"{_default_suffix}azure_handler"] = {
            "level": lowest_level_to_log,
            "class": f"{AzureLogHandler.__module__}.{AzureLogHandler.__name__}",
            "connection_string": azure_insights_connection_string,
            "formatter": PHAC_HELPER_JSON_FORMATTER_KEY,
        }

    slack_webhook_url = get_logging_env_value("SLACK_WEBHOOK_URL")
    if slack_webhook_url is not None:
        # pylint: disable=ungrouped-imports
        from phac_aspc.django.helpers.logging.json_post_handlers import (
            SlackWebhookHandler,
        )

        class NoisyLoggerFilter:
            def filter(self, record):
                noisy_loggers = ["django.security.DisallowedHost"]

                if record.module in noisy_loggers:
                    return 0

                return 1

        noisy_logger_key = f"{_default_suffix}slack_webhook_handler_filter"

        additional_filter_configs[noisy_logger_key] = {
            "()": NoisyLoggerFilter,
        }

        additional_handler_configs[
            f"{_default_suffix}slack_webhook_handler"
        ] = {
            "level": "ERROR",
            "class": f"{SlackWebhookHandler.__module__}.{SlackWebhookHandler.__name__}",
            "url": slack_webhook_url,
            "formatter": PHAC_HELPER_PRETTY_JSON_FORMATTER_KEY,
            "filters": [noisy_logger_key],
        }

    configure_uniform_std_lib_and_structlog_logging(
        lowest_level_to_log=lowest_level_to_log,
        mute_console_handler=get_logging_env_value("MUTE_CONSOLE_HANDLER"),
        additional_handler_configs=additional_handler_configs,
        console_handler_formatter_key=(
            PHAC_HELPER_CONSOLE_FORMATTER_KEY
            if get_logging_env_value("PRETTY_FORMAT_CONSOLE_LOGS")
            else PHAC_HELPER_JSON_FORMATTER_KEY
        ),
        additional_filter_configs=additional_filter_configs,
    )



"""Handlers for JSON POST requests, for use alongside the PHAC helpers logging configuration"""

import logging.config
from abc import ABCMeta, abstractmethod

try:
    import requests
except (ImportError, ModuleNotFoundError) as exc:
    raise ImportError(
        "The `requests` package is required for use of the PHAC helpers JSON post"
        + "handlers. You must install this dependency your self."
    ) from exc


class AbstractJSONPostHandler(logging.Handler, metaclass=ABCMeta):
    """
    Handler for sending JSON formatted logs to an arbitrary endpoints. Subclasses must
    implement a get_json_from_record method to convert log records to match the target
    endpoint's API.

    Initialization requires a URL to post to, and also accepts an optional flag to
    control whether post attempts fail silently.

    Even if not failing silently, the handler will not try to post it's own error
    logs, to avoid potential failure loops when the target endpoint is unreachable.
    """

    def __init__(self, url: str, fail_silent: bool = False):
        super().__init__()
        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}"
        )

        self.url = url
        self.fail_silent = fail_silent

    def emit(self, record):
        is_own_error_log = record.name == self.logger.name and (
            record.levelname in ("ERROR", "CRITICAL")
        )

        if not is_own_error_log:
            try:
                response = requests.post(
                    self.url,
                    json=self.get_json_from_record(record),
                    timeout=1,
                )

                response.raise_for_status()

            except requests.RequestException as exception:
                if not self.fail_silent:
                    self.logger.error(
                        '%s\'s logging request to URL "%s" failed',
                        self.__class__.__name__,
                        self.url,
                        exc_info=exception,
                    )

    @abstractmethod
    def get_json_from_record(self, record):
        pass


class SlackWebhookHandler(AbstractJSONPostHandler):
    """Trivial Slack Webhook JSON post handler"""

    def get_json_from_record(self, record):
        return {"text": self.format(record)}



from phac_aspc.django.helpers.jinja_utils import include_from_dtl
from phac_aspc.jinja.registry import registry

registry.add_global("include_from_dtl", include_from_dtl)


def environment(**options):
    env = registry.get_environment(**options)
    return env



"""Utility to get language code from lang strings with locales"""

from django.utils.translation import get_language, get_language_info


def get_language_code(lang=None):
    """Translate strings like fr-ca, en-ca to fr and en respectively
    If lang is not specified the current active language will be used.
    """
    if not lang:
        lang = get_language()
    try:
        return get_language_info(lang)["code"]
    except TypeError:
        return "en"



# This module is a wrapper around the django-rules package
from unittest.mock import patch

try:
    import rules
    from rules import add_rule, predicate
except ImportError:
    raise ImportError("rules package is not installed")


class NonExistentRuleException(Exception):
    pass


# this is the "private" version, for mocking purposes
def _test_rule(name, user=None, obj=None):
    if not rules.rule_exists(name):
        raise NonExistentRuleException(f"rule {name} does not exist")

    return rules.test_rule(name, user, obj)


def test_rule(*args, **kwargs):
    return _test_rule(*args, **kwargs)


def auto_rule(fn):
    """
    use as decorator, e.g.

    @auto_rule
    def rule_name(user, obj):
        ...


    is shorthand for

    add_rule("rule_name", rule_name_func)

    """
    pred = predicate(fn)
    add_rule(fn.__name__, pred)
    return pred


class patch_rules:
    """
    usage: with mock_rules(can_access_foo=False):
      assert test_client.get(some_view_that_uses_patched_rules).status_code == 403
    """

    def rule_mocker(self, **rule_stubs):
        def exec_rule(rule_name, user=None, obj=None):
            if rule_name in rule_stubs:
                return rule_stubs[rule_name]

            return self.actual_rule_func(rule_name, user, obj)

        return exec_rule

    def __init__(self, **rule_stubs):
        self.actual_rule_func = _test_rule
        self._patch = patch(
            "phac_aspc.rules._test_rule", self.rule_mocker(**rule_stubs)
        )

    def __enter__(self):
        return self._patch.__enter__()

    def __exit__(self, *excp):
        return self._patch.__exit__(*excp)



"""This file contains a utility for detecting when the django is running inside
of a test framework environment
"""

import sys


def is_running_tests():
    """Detect if the app process was launched by a test-runner command"""
    return "test" in sys.argv or any("pytest" in arg for arg in sys.argv)



from xml.etree import ElementTree

from django import template
from django.contrib.staticfiles import finders
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag
def phac_aspc_inline_svg(static_file_path=None, **kwargs):
    """Inlines a SVG icon from linkcube/src/core/assets/templates/core.assets/icon

    Based on https://stackoverflow.com/a/68802035

    Example usage:
        {% phac_aspc_inline_svg "phac_aspc_helpers/phac_logos/en.svg" height="2rem" %}
    Parameter: static_file_path
        Path of the svg to be inlined; resolved relative to the static root(s) of installed apps.
    Parameter: kwargs
        Additional kwargs are treated as attribute:value pairs and set on the svg's root node.
    Returns:
        stringified XML to be inlined, i.e.:
        '<svg class="..." height="...">...</svg>'
    """
    ElementTree.register_namespace("", "http://www.w3.org/2000/svg")

    svg_root = ElementTree.parse(finders.find(static_file_path)).getroot()

    for attribute, value in kwargs.items():
        svg_root.set(attribute, value)

    svg_string = ElementTree.tostring(
        svg_root, encoding="unicode", method="html"
    )

    return mark_safe(svg_string)



from django.urls import include, path, re_path


def view_for_template(template_name):
    def view_func(request):
        from django.shortcuts import render

        return render(request, template_name)

    return view_func


urlpatterns = [
    path(
        "with_language_tag/",
        view_for_template("language_tag_example.jinja2"),
        name="with_language_tag",
    ),
]



"""This file contains utilities for asserting that consuming Django apps meet the
library settings requirements.
"""

import inspect

from django.core import checks


def trigger_configuration_warning(msg, **kwargs):
    """Trigger a configuration error programmatically"""

    def phac_aspc_conf_warning(
        app_configs, **conf_kwargs
    ):  # pylint: disable=unused-argument
        return [checks.Warning(msg, **kwargs)]

    checks.register(phac_aspc_conf_warning)


def warn_and_remove(items, dest_list):
    """If the items intersects with dest_list, raise a configuration warning and
    remove it from the list returned to prevent errors."""
    ret = []
    context = inspect.stack()[1].function
    hint = "You should remove it from your project to ensure proper ordering."

    for item in items:
        if item in dest_list:
            trigger_configuration_warning(
                f"{item} is already defined in the provided list.",
                id=context,
                hint=hint,
            )
            hint = None
        else:
            ret.append(item)
    return ret


def configure_apps(app_list):
    """Return an application list which includes the apps required by this
    library"""

    # Modules that read global state are best deffered to call time rather than module-load
    # pylint: disable=import-outside-toplevel
    from phac_aspc.django.settings.logging_env import get_logging_env_value

    prefix_list = warn_and_remove(["modeltranslation", "axes"], app_list)

    logging_app = (
        ["django_structlog"]
        if get_logging_env_value("USE_HELPERS_CONFIG")
        else []
    )

    suffix_list = warn_and_remove(
        [
            "phac_aspc.django.helpers",
            "rules.apps.AutodiscoverRulesConfig",
            *logging_app,
        ],
        app_list,
    )

    return prefix_list + app_list + suffix_list


def configure_authentication_backends(backend_list):
    """Return the authentication backend list includes those required by this
    library.

    By default importing the settings will automatically configure the backend,
    however if you want to customize the authentication backend used by your
    project, you can use this method to ensure proper configuration."""

    # Modules that read global state are best deffered to call time rather than module-load
    # pylint: disable=import-outside-toplevel
    from phac_aspc.django.settings.security_env import get_oauth_env_value

    oauth_backend = (
        [get_oauth_env_value("USE_BACKEND")]
        if get_oauth_env_value("PROVIDER")
        and get_oauth_env_value("USE_BACKEND")
        else []
    )

    prefix_backends = warn_and_remove(
        ["axes.backends.AxesStandaloneBackend"] + oauth_backend, backend_list
    )
    return prefix_backends + backend_list


def configure_middleware(middleware_list):
    """Return the list of middleware configured for this library"""
    # Modules that read global state are best deffered to call time rather than module-load
    # pylint: disable=import-outside-toplevel
    from phac_aspc.django.settings.logging_env import get_logging_env_value

    logging_middleware = (
        ["django_structlog.middlewares.RequestMiddleware"]
        if get_logging_env_value("USE_HELPERS_CONFIG")
        else []
    )

    prefix = warn_and_remove(
        [
            "axes.middleware.AxesMiddleware",
            "django.middleware.locale.LocaleMiddleware",
        ]
        + logging_middleware,
        middleware_list,
    )
    return prefix + middleware_list



"""This file contains utilities for accessing env values used by consuming libraries
to configure PHAC helpers' behaviour
"""

import inspect
import os
from importlib.util import find_spec

import environ

PHAC_ENV_PREFIX = "PHAC_ASPC_"


def get_env_value(env, key, prefix=PHAC_ENV_PREFIX):
    """Returns the value for the prefixed key from the provided env, unless that
    key exists in the Django settings in which case the settings value is used
    """
    # Modules that read global state are best deffered to call time rather than module-load
    from django.conf import settings  # pylint: disable=import-outside-toplevel

    prefixed_key = f"{prefix}{key}"

    return getattr(
        settings,
        prefixed_key,
        env(prefixed_key),
    )


def find_env_file(path):
    """Traverse directories backwards starting at `path`, returns the path of the
    first .env file found, or `None` otherwise.
    """

    # look for .env file in provided path
    filename = os.path.join(path, ".env")
    if os.path.isfile(filename):
        return filename

    # when called on the root dir, os.path.dirname returns the root dir2
    parent = os.path.dirname(path)

    # recurse backwards till we reach the root
    # (doesn't sit right, but there's no non-hacky way to know where the
    # consuming repo's root is; other env packages like dotenv also
    # default to searching backwards till they hit root)
    if parent and parent != path:
        return find_env_file(parent)

    return None


def get_env(prefix=PHAC_ENV_PREFIX, **conf):
    """Return django-environ configured with the provided values and
    using the prefix.

    `prefix` can be used to change the environment variable prefix that is added
    to the beginning on the variables defined in conf. By default this value is
    `PHAC_ASPC_`.

    `conf` is a dictionary used to generate the scheme for django-environ.

    Will attempt to find and load from a .env, starting in the directory of the consuming
    application's `DJANGO_SETTINGS_MODULE` and traveling back towards root. Continues
    even if no .env is found, with every env var taking it's default from `conf`
    instead.

    See https://django-environ.readthedocs.io/en/latest/api.html#environ.Env for
    additional information on the scheme.
    """

    scheme = {}
    for name, values in conf.items():
        scheme[f"{prefix}{name}"] = values

    env = environ.Env(**scheme)

    settings_path_of_consuming_app = os.path.dirname(
        find_spec(os.getenv("DJANGO_SETTINGS_MODULE")).origin
    )
    nearest_ancestor_dot_env = find_env_file(settings_path_of_consuming_app)
    if nearest_ancestor_dot_env and os.path.isfile(nearest_ancestor_dot_env):
        environ.Env.read_env(nearest_ancestor_dot_env)

    return env


def global_from_env(prefix=PHAC_ENV_PREFIX, **conf):
    """Create named global variables based on the provided environment variable
    scheme.  Variables defined in the scheme will be inserted into the calling
    module's globals and prefixed with `PHAC_ASPC_` when fetching the
    environment variable.

    prefix can be used to change the environment variable prefix that is added
    to the beginning on the variables defined in conf.  By default this value is
    `PHAC_ASPC_`.

    conf is a dictionary used to generate the scheme for django-environ.
    """

    mod = inspect.getmodule(inspect.stack()[1][0])
    env = get_env(prefix, **conf)

    for name in conf:
        setattr(mod, name, get_env_value(env, name, prefix))



"""
Enables support for executing code when all django apps are ready.
"""

# Define a singleton that will contain the functions to be fired
# on ready.
ready_hooks = []


def execute_when_ready(func):
    """
    Registers a function that will be delayed until all django apps are ready.
    """
    ready_hooks.append(func)


def process_ready_hooks():
    """
    Process the registered hooks.

    This should occur after the models are loaded by Django, for example from
    the AppConfig's ready function.
    """
    for hook in ready_hooks:
        hook()

    ready_hooks.clear()



"""Setup URLs for views related to WET"""

from django.urls import path

from ..views import wet

urlpatterns = [
    path(
        "phac-aspc/helpers/session",
        wet.session,
        name="phac_aspc_helpers_session",
    )
]



# pylint: disable=missing-function-docstring, missing-class-docstring, too-few-public-methods, too-many-ancestors
"""
This contains custom fields that should be used in place of the default Django fields.

The django migration engines serializes all fields and their attributes.
There are many attributes that are not relevant to the database schema.
The fields below override serialization and strip irrelevant attributes.

These fields strip attributes from serialization
so we don't get extra migrations when we change verbose names or related_names

"""

from django.db import models

from . import comma_separated_field

migration_ignored_attrs = [
    "help_text",
    "is_searchable",
    "description",
    "verbose_name",
    "verbose_name_plural",
    "ordering",
    "validators",
    "blank",
    "extra_options",
    "related_name",
    "choices",
]


class SearchableFieldMixin:
    def __init__(self, *args, is_searchable=False, **kwargs):
        self.is_searchable = is_searchable
        super().__init__(*args, **kwargs)


class DescriptionMixin:
    def __init__(self, *args, extra_options=None, description="", **kwargs):
        self.extra_options = extra_options or {}
        self.custom_description = description
        super().__init__(*args, **kwargs)

    @property
    def description(self):
        return self.custom_description

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        for attr in migration_ignored_attrs:
            kwargs.pop(attr, None)
        return name, path, args, kwargs


# pylint: disable=abstract-method
class ManyToManyField(DescriptionMixin, models.ManyToManyField):
    pass


class EmptyToNullMixin:
    """
    Replace NULL values with empty string ""
    So there's only one way to represent an empty string
    """

    def __init__(self, *args, null_to_empty=False, **kwargs):
        if null_to_empty:
            bad_keys = [
                "choices",
                "null",
                "default",
                "dbindex",
                "primary_key",
                "unique",
            ]
            if any(key in kwargs for key in bad_keys):
                raise TypeError(
                    " Improper use of fields, you probably want a NULL-able column"
                )

            super().__init__(
                *args,
                null=False,
                default="",
                **kwargs,
            )

        else:
            super().__init__(*args, **kwargs)


class CharField(
    SearchableFieldMixin, DescriptionMixin, EmptyToNullMixin, models.CharField
):
    pass


class TextField(
    SearchableFieldMixin, DescriptionMixin, EmptyToNullMixin, models.TextField
):
    def deconstruct(self):
        # according to django docs, TextField char-limits are not enforced at the DB level
        name, path, args, kwargs = super().deconstruct()
        kwargs.pop("max_length", None)

        return name, path, args, kwargs


# pylint: disable=abstract-method
class ForeignKey(DescriptionMixin, models.ForeignKey):
    pass


choices_by_type = {
    "yes_or_no": ((True, "yes"), (False, "no")),
    "true_or_false": ((True, "true"), (False, "false")),
}


class BooleanField(DescriptionMixin, models.BooleanField):
    def __init__(self, *args, **kwargs):
        if "choices" not in kwargs:
            # default to True/False
            choice_type = kwargs.pop("choice_type", "yes_or_no")
            kwargs["choices"] = choices_by_type[choice_type]

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs.pop("choices", None)
        kwargs.pop("choice_type", None)

        return name, path, args, kwargs


class DecimalField(DescriptionMixin, models.DecimalField):
    pass


class FloatField(DescriptionMixin, models.FloatField):
    pass


class IntegerField(DescriptionMixin, models.IntegerField):
    pass


class PositiveIntegerField(DescriptionMixin, models.PositiveIntegerField):
    pass


class SlugField(DescriptionMixin, models.SlugField):
    pass


# pylint: disable=abstract-method
class OneToOneField(DescriptionMixin, models.OneToOneField):
    pass


class AutoField(DescriptionMixin, models.AutoField):
    pass


class BigAutoField(DescriptionMixin, models.BigAutoField):
    pass


class DateTimeField(DescriptionMixin, models.DateTimeField):
    pass


class DateField(DescriptionMixin, models.DateField):
    pass


class EmailField(DescriptionMixin, models.EmailField):
    pass


class URLField(DescriptionMixin, models.URLField):
    pass


class ImageField(DescriptionMixin, models.ImageField):
    pass


class FileField(DescriptionMixin, models.FileField):
    pass


class CommaSeparatedCharField(
    DescriptionMixin, comma_separated_field.CommaSeparatedCharField
):
    pass


class CommaSeparatedTextField(
    DescriptionMixin, comma_separated_field.CommaSeparatedTextField
):
    pass



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def assert_both_dtl_and_jinja_configured():
    are_both_dtl_and_jinja_configured = map(
        lambda template_backend: next(
            filter(
                lambda template_config: template_config["BACKEND"]
                == template_backend,
                settings.TEMPLATES,
            ),
            False,
        ),
        [
            "django.template.backends.django.DjangoTemplates",
            "django.template.backends.jinja2.Jinja2",
        ],
    )

    if not are_both_dtl_and_jinja_configured:
        raise ImproperlyConfigured(
            "settings.TEMPLATES must include both the DjangoTemplate and Jinja2 backends"
        )

    return are_both_dtl_and_jinja_configured



"""Logging configuration env var configs and getters"""

from phac_aspc.django.settings.utils.env_utils import (
    PHAC_ENV_PREFIX,
    get_env,
    get_env_value,
)
from phac_aspc.django.settings.utils.is_running_tests import is_running_tests

LOGGING_ENV_PREFIX = f"{PHAC_ENV_PREFIX}LOGGING_"

logging_env = get_env(
    prefix=LOGGING_ENV_PREFIX,
    USE_HELPERS_CONFIG=(bool, False),
    LOWEST_LEVEL=(str, "INFO"),
    MUTE_CONSOLE_HANDLER=(bool, is_running_tests()),
    PRETTY_FORMAT_CONSOLE_LOGS=(bool, False),
    AZURE_INSIGHTS_CONNECTION_STRING=(str, None),
    SLACK_WEBHOOK_URL=(str, None),
)


def get_logging_env_value(key):
    return get_env_value(logging_env, key, prefix=LOGGING_ENV_PREFIX)



from django import template
from django.template.loader import get_template

from phac_aspc.django.helpers.jinja_dtl_interop_utils import (
    assert_both_dtl_and_jinja_configured,
)

register = template.Library()


@register.simple_tag(takes_context=True)
def phac_aspc_include_from_jinja(context, template_name):
    """
    Renders a Jinja2 template inside a DTL template, using the DTL template's context.
    """
    assert_both_dtl_and_jinja_configured()

    jinja_template = get_template(template_name)

    return jinja_template.render(context.flatten())



# pylint: disable=missing-function-docstring, missing-class-docstring
"""
utilities for writing excel data using openpyxl
"""

import csv
import re

from django.core.paginator import Paginator
from django.db.models import QuerySet
from django.http import HttpResponse
from django.utils.functional import Promise
from django.utils.safestring import SafeString
from django.views import View

from openpyxl.cell import WriteOnlyCell
from openpyxl.utils import get_column_letter

try:
    import openpyxl
    from openpyxl.cell.cell import ILLEGAL_CHARACTERS_RE
    from openpyxl.utils import escape as escapeSvc
except (ImportError, ModuleNotFoundError) as exc:
    raise ImportError(
        "must install openpyxl==3.0.10 to use excel helpers"
    ) from exc

# Note that OPENPXYL starts columns and rows at index 1.


GLOBAL_MAX_COL_LEN = 5000


def escape_for_xlsx(text):
    """
    escape a string for excel
    """
    # pylint: disable=unidiomatic-typecheck
    return escapeSvc.escape(text) if isinstance(text, str) else text


def create_field_val_getter(field_name):
    """
    returns a function that gets the value of a field for a django model instance
    """
    return lambda record: record.serializable_value(field_name)


def write_queryset_to_sheet(workbook, queryset):
    """
    iterates over queryset and puts all fields the workbook in a sheet
    - picks scalars and FK-IDs based on model
    - picks sheet name based on model
    - warning: uses queryset.iterator, prefetches are ignored
    """

    ModelToSheetWriter(workbook=workbook, iterator=queryset).write()


def page_queryset(queryset, per_page=1000):
    """
    While in a perfect world you would use queryset.iterator()
    but there is an issue that it fails to
    load any prefetch_related() fields specified.
    Paginator() can mimic the same functionality

    https://github.com/django-import-export/django-import-export/issues/774#issuecomment-449064652
    """

    # pylint: disable=protected-access
    if queryset._prefetch_related_lookups:
        if not queryset.query.order_by:
            # Paginator() throws a warning if there is no sorting attached to the queryset
            queryset = queryset.order_by("pk")
        paginator = Paginator(queryset, per_page)
        for index in range(paginator.num_pages):
            yield from paginator.get_page(index + 1)
    else:
        yield from queryset.iterator(chunk_size=per_page)


class Column:
    """
    Base class to write columns in a sheet
    """

    style = None
    column_width = None

    def __init__(self, style=None, column_width=None):
        self.style = style
        self.column_width = column_width

    def get_header(self):
        """return a header string for the column"""
        raise NotImplementedError()

    def get_value(self, record):
        """return this column's intended cell value for a record"""
        raise NotImplementedError()

    def serialize_value(self, value):
        return serialize_value(value)

    def get_serialized_value(self, record):
        return self.serialize_value(self.get_value(record))

    def get_style(self):
        return self.style

    def get_column_width(self):
        return self.column_width


class ModelColumn(Column):
    """
    shorthand for defining a column that writes a scalar model field (non foreign-key)
    """

    def __init__(self, model_cls, field_name, header_value=None, **kwargs):
        super().__init__(**kwargs)
        self.header_value = header_value
        self.model_cls = model_cls
        self.field_name = field_name
        self.get_val = create_field_val_getter(field_name)

    def get_header(self):
        # pylint: disable=protected-access
        header_value = (
            self.header_value
            or self.model_cls._meta.get_field(self.field_name).verbose_name
        )
        return escape_for_xlsx(header_value)

    def get_value(self, record):
        return self.get_val(record)


class ChoiceColumn(Column):
    def __init__(self, model_cls, field_name, header_value=None, **kwargs):
        super().__init__(**kwargs)
        self.header_value = header_value
        self.model_cls = model_cls
        self.field_name = field_name

    def get_header(self):
        # pylint: disable=protected-access
        header_value = (
            self.header_value
            or self.model_cls._meta.get_field(self.field_name).verbose_name
        )
        return escape_for_xlsx(header_value)

    def get_value(self, record):
        method_name = "get_" + self.field_name + "_display"
        return getattr(record, method_name)()


class CustomColumn(Column):
    def __init__(self, header, get_val, **kwargs):
        super().__init__(**kwargs)
        self.header = header
        self.get_val = get_val

    def get_header(self):
        return self.header

    def get_value(self, record):
        return self.get_val(record)


class ManyToManyColumn(Column):
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        model: type,
        field_name: str,
        get_related_str: callable = None,
        delimiter: str = ", ",
        header=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model = model
        self.field_name = field_name
        self.header = header
        if get_related_str:
            self.get_related_str = get_related_str
        else:
            self.get_related_str = lambda x, *args: str(x)

        self.delimiter = delimiter

    def get_header(self):
        if self.header:
            return self.header
        # pylint: disable=protected-access
        return self.model._meta.get_field(self.field_name).verbose_name

    def get_value(self, record):
        related_records = list(getattr(record, self.field_name).all())
        return self.delimiter.join(
            [self.get_related_str(x) for x in related_records]
        )


class WriterConfigException(Exception):
    pass


class AbstractWriter:
    iterator = None
    columns = None

    def __init__(self, iterator=None, columns=None):
        if iterator is not None:
            self.iterator = iterator

        if columns is not None:
            self.columns = columns

    def get_iterator(self):
        if self.iterator is None:
            raise NotImplementedError(
                "must set iterator in init, class or override get_iterator()"
            )
        return self.iterator

    def get_header_row(self):
        # return [escape_for_xlsx(name) for name, _ in self.get_column_configs()]
        return [
            serialize_value(col.get_header())
            for col in self.get_column_configs()
        ]

    def get_column_configs(self):
        """
        Legacy (now internal) API
        override get_columns(), columns attr or columns init kwarg instead
        """
        return self.get_columns()

    def get_columns(self):
        if self.columns is not None:
            return self.columns

        raise NotImplementedError(
            "must set columns in init, class or override get_column_configs()"
        )

    def write(self):
        raise NotImplementedError()


class AbstractModelWriter(AbstractWriter):
    # pylint: disable=W0223

    def __init__(self, **kwargs):
        if "iterator" in kwargs:
            raise WriterConfigException(
                "don't use iterator kwarg, use the queryset kwarg, "
                "class attr or override get_queryset()"
            )
        if "queryset" in kwargs:
            self.queryset = kwargs.pop("queryset")

        super().__init__(**kwargs)

    def get_iterator(self):
        return page_queryset(self.get_queryset())

    def get_queryset(self):
        """
        This is just here for legacy sake,
        some consumer subclasses have overriden this method
        """
        try:
            return self.queryset
        except AttributeError as e:
            raise WriterConfigException(
                "Must define queryset attr, kwarg or override get_queryset()"
            ) from e

    def get_columns(self):
        try:
            super_cols = super().get_columns()
        except NotImplementedError:
            super_cols = None

        if super_cols is not None:
            return super_cols

        model = self.get_queryset().model

        fields_to_write = model._meta.fields
        return [ModelColumn(model, field.name) for field in fields_to_write]


class AbstractCsvWriter(AbstractWriter):
    # pylint: disable=W0223

    def __init__(self, buffer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.buffer = buffer

    def write(self):
        writer = csv.writer(self.buffer)
        writer.writerow(self.get_header_row())

        iterator = self.get_iterator()
        for record in iterator:
            csv_row = []
            for col in self.get_column_configs():
                csv_val = col.get_serialized_value(record)
                csv_row.append(csv_val)

            writer.writerow(csv_row)


class ModelToCsvWriter(AbstractCsvWriter, AbstractModelWriter):
    pass


class AbstractSheetWriter(AbstractWriter):
    # pylint: disable=W0223

    sheet_name = None
    header_style = None

    def __init__(self, *args, workbook=None, sheet_name=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.workbook = workbook

        if sheet_name:
            self.sheet_name = sheet_name

    def get_sheet_name(self):
        if not self.sheet_name:
            raise NotImplementedError(
                "must override get_sheet_name() or pass a sheet_name"
            )

        return self.sheet_name

    def get_header_style(self):
        """
        To be overwritten by child classes when a style is to be applied to the
        header row.  If overwritten, should return a Style object.
        """
        return self.header_style

    def get_column_widths(self):
        """
        Returns an array of column widths as set in the get_column_configs.
        """
        return [col.get_column_width() for col in self.get_column_configs()]

    def write(self):
        worksheet = self.workbook.create_sheet(title=self.get_sheet_name())

        for col_index, column_width in enumerate(self.get_column_widths()):
            if column_width is not None:
                column_letter = get_column_letter(col_index + 1)
                worksheet.column_dimensions[column_letter].width = column_width

        header_labels = self.get_header_row()
        # pylint: disable=E1128
        header_style = self.get_header_style()

        header_row = []
        for label in header_labels:
            cell = WriteOnlyCell(worksheet, value=label)

            if header_style:
                cell.style = header_style
            header_row.append(cell)

        worksheet.append(header_row)

        iterator = self.get_iterator()

        for record in iterator:
            xl_row = []
            for col in self.get_column_configs():
                xl_val = col.get_serialized_value(record)
                cell = WriteOnlyCell(worksheet, value=xl_val)
                style = col.get_style()
                if style:
                    cell.style = style
                xl_row.append(cell)

            worksheet.append(xl_row)


class ModelToSheetWriter(AbstractSheetWriter, AbstractModelWriter):
    def get_sheet_name(self):
        try:
            return super().get_sheet_name()
        except NotImplementedError:
            return get_default_sheet_name_for_qs(self.get_queryset())


def serialize_value(value):
    if value is None:
        write_val = ""
    else:
        if value is True:
            write_val = 1
        elif value is False:
            write_val = 0
        elif isinstance(value, list):
            write_val = str(value)
        elif isinstance(value, (Promise, SafeString)):
            # tm values don't play well with excel
            write_val = value + ""
        else:
            write_val = value

    xl_val = escape_for_xlsx(write_val)
    if next(ILLEGAL_CHARACTERS_RE.finditer(str(value)), None):
        xl_val = re.sub(ILLEGAL_CHARACTERS_RE, "", str(xl_val))

    return xl_val


def get_default_sheet_name_for_qs(queryset):
    """
    excel only supports up to 31 characters in a worksheet name
    """
    # pylint: disable=protected-access
    model_name = queryset.model._meta.verbose_name
    return f"{model_name[:30]}"


class BaseAbstractExportView(View):
    """
    Must implement get_iterator(), get_queryset() OR assign self.queryset

    """

    filename = None
    sheetwriter_class = None

    def get_filename(self):
        if not getattr(self, "filename", None):
            raise NotImplementedError(
                "Must define filename attr or override get_filename()"
            )

    def _get_iterable(self):
        if hasattr(self, "queryset"):
            assert isinstance(
                self.queryset, QuerySet
            ), "queryset must be a QuerySet"
            return self.queryset
        if hasattr(self, "iterator"):
            return self.iterator

        if hasattr(self, "get_queryset"):
            qs = self.get_queryset()
            assert isinstance(qs, QuerySet), "queryset must be a QuerySet"
            return qs

        if hasattr(self, "get_iterator"):
            return self.get_iterator()

        return None

    def get_sheetwriter_class(self) -> type:
        if not self.sheetwriter_class:
            raise NotImplementedError(
                "Must define sheetwriter_class attr or override get_sheetwriter_class()"
            )
        return self.sheetwriter_class


class AbstractExportView(BaseAbstractExportView):
    filename = "export.xlsx"

    def get(self, request, *args, **kwargs):
        wb = openpyxl.Workbook(write_only=True)
        WriterCls = self.get_sheetwriter_class()

        iterable = self._get_iterable()
        if iterable is None:
            writer = WriterCls(workbook=wb)
        elif issubclass(WriterCls, AbstractModelWriter):
            if not isinstance(iterable, QuerySet):
                raise WriterConfigException("iterable must be a QuerySet")
            writer = WriterCls(
                workbook=wb,
                queryset=iterable,
            )
        else:
            writer = WriterCls(workbook=wb, iterator=iterable)

        writer.write()
        response = HttpResponse(
            headers={"Content-Type": "application/vnd.ms-excel"}
        )
        response["Content-Disposition"] = (
            f"attachment; filename={self.get_filename()}"
        )
        wb.save(response)
        return response


class AbstractCsvExportView(AbstractExportView):
    filename = "export.csv"

    def get_writer_class(self):
        try:
            return self.writer_class
        except AttributeError as e:
            raise NotImplementedError(
                "Must define writer_class attr or override get_writer_class()"
            ) from e

    def get(self, request, *args, **kwargs):
        response = HttpResponse(
            content_type="text/csv; charset=utf-8",
            headers={
                "Content-Disposition": f"attachment; filename={self.get_filename()}"
            },
        )

        WriterCls = self.get_writer_class()
        iterable = self._get_iterable()
        if iterable is None:
            writer = WriterCls(buffer=response)

        elif issubclass(WriterCls, AbstractModelWriter):
            if not isinstance(iterable, QuerySet):
                raise WriterConfigException("iterable must be a QuerySet")
            writer = WriterCls(
                queryset=iterable,
                buffer=response,
            )
        else:
            writer = WriterCls(buffer=response, iterator=iterable)

        writer.write()
        return response



from phac_aspc.rules import (
    add_rule,
    auto_rule,
    patch_rules,
)
from phac_aspc.rules import (
    test_rule as func_test_rule,  # must be imported in form that doesn't start with test_; otherwise pytest will try to run it as a test
)


def test_rules():
    @auto_rule
    def has_foo_access(user, obj):
        return True

    # also test manually added rule
    def has_bar_access_func(user, obj):
        return False

    add_rule("has_bar_access", has_bar_access_func)

    # test_rule works with 1, 2, or 3 arguments
    assert func_test_rule("has_foo_access")
    assert func_test_rule("has_foo_access", 1)
    assert func_test_rule("has_foo_access", 1, 1)
    assert not func_test_rule("has_bar_access")

    with patch_rules(has_foo_access=False, has_bar_access=True):
        assert func_test_rule("has_bar_access")
        assert not func_test_rule("has_foo_access")

    # rules are returned to normal after patch_rules
    assert not func_test_rule("has_bar_access")
    assert func_test_rule("has_foo_access")



from phac_aspc.vanilla import classproperty, flatten, group_by


def test_group_by():
    assert group_by([1, 2, 3, 4, 5, 6], lambda x: x % 2) == {
        0: [2, 4, 6],
        1: [1, 3, 5],
    }
    assert group_by((1, 2, 3, 4, 5, 6), lambda x: x % 2) == {
        0: [2, 4, 6],
        1: [1, 3, 5],
    }


def test_flatten():
    assert flatten([[1], [2, 3], [4, 5, 6]]) == [1, 2, 3, 4, 5, 6]
    assert flatten(([1], [2, 3], (4, 5, 6))) == [1, 2, 3, 4, 5, 6]


def test_class_property():
    class MyClass:
        @classproperty
        def my_class_property(cls):  # pylint: disable=no-self-argument
            return 1

    class MyChildClass:
        @classproperty
        def my_class_property(cls):  # pylint: disable=no-self-argument
            return 2

    # pylint: disable=comparison-with-callable
    assert MyClass.my_class_property == 1
    # pylint: disable=comparison-with-callable
    assert MyChildClass.my_class_property == 2



"""Decorators related to django admin"""

from django.contrib.admin import ModelAdmin, TabularInline, site

from modeltranslation.admin import TranslationAdmin

from phac_aspc.django.helpers.ready import execute_when_ready


class add_admin:  # pylint: disable=invalid-name,too-few-public-methods
    """
    Add a model Django's administration interface.

    You can add a list of inlines models using the `inlines` property,
    or any other option supported by `ModelAdmin` via the `admin_options`
    property.

    Usage:
        >>> from django.db import models
        >>> from phac_aspc.django.admin.decorators import add_admin
        >>>
        >>> @add_admin()
        >>> class Person(models.Model):
        >>>     ...
        >>>

    """

    def __init__(self, inlines=None, admin_options=None):
        self.ao = admin_options
        self.inlines = inlines

    def __call__(self, cls):
        def register_model():
            opts = self.ao or {}
            opts["inlines"] = []
            if self.inlines is not None:
                for model in self.inlines:
                    opts["inlines"].append(
                        type(
                            'f"{cls.__name__}_{model}InlineAdmin"',
                            (TabularInline,),
                            {"model": model},
                        )
                    )

            base_class = (
                TranslationAdmin
                if hasattr(cls, "__is_translated_model__")
                else ModelAdmin
            )

            site.register(
                cls, type('f"{cls.__name__}Admin"', (base_class,), opts)
            )

        execute_when_ready(register_model)
        return cls



"""Recommended localization settings"""

from django.utils.translation import gettext_lazy as _

from .localization_env import localization_env_config
from .utils import global_from_env

LANGUAGES = (
    ("fr-ca", _("French")),
    ("en-ca", _("English")),
)

global_from_env(**localization_env_config)



from django.template.loader import get_template

from jinja2 import pass_context

from phac_aspc.django.helpers.jinja_dtl_interop_utils import (
    assert_both_dtl_and_jinja_configured,
)


@pass_context
def include_from_dtl(context, template_name):
    """
    Renders a DTL template inside a Jinja2 template, using the Jinja2 template's context.
    """
    assert_both_dtl_and_jinja_configured()

    dtl_template = get_template(template_name)

    return dtl_template.render(context.get_all())



"""Related to Authentication"""

from django import template
from django.template import loader

from phac_aspc.django.settings.security_env import get_oauth_env_value

register = template.Library()


@register.simple_tag()
def phac_aspc_auth_signin_microsoft_button(query_params=""):
    """Returns a signin button using the microsoft design"""

    return (
        loader.get_template(
            "phac_aspc/helpers/auth/buttons/microsoft.html"
        ).render({"query_params": query_params})
        if get_oauth_env_value("PROVIDER")
        else ""
    )



"""Example authentication backend used by oauth login flow"""

from typing import Any

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.base_user import (  # pylint: disable=unused-import
    AbstractBaseUser,
)
from django.http.request import HttpRequest


class PhacAspcOAuthBackend(BaseBackend):
    """Authentication backend that creates a user using only the oid and email"""

    def _sync_user(self, user, user_info, force=False):
        email = user_info["email"] if "email" in user_info else ""
        if force or (email not in ("", user.email)):
            user.email = email
            user.save()

    def authenticate(
        self,
        request: HttpRequest,
        user_info: "dict | None" = None,
        **kwargs: Any,
    ) -> "AbstractBaseUser | None":
        if user_info is not None:
            user_model = get_user_model()
            try:
                user = user_model.objects.get(username=user_info["oid"])
                self._sync_user(user, user_info)
            except user_model.DoesNotExist:
                user = user_model(username=user_info["oid"])
                self._sync_user(user, user_info, True)
            return user
        return None

    def get_user(self, user_id):
        user_model = get_user_model()
        try:
            return user_model.objects.get(pk=user_id)
        except user_model.DoesNotExist:
            return None



"""Utilities for use alongside the PHAC helpers logging configuration"""

from typing import Any, Dict

from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed

import structlog


def add_fields_to_all_logs_for_current_request(context_dict: Dict[str, Any]):
    """Binds structlog contextvars to all subsequent logs for the request currently being processed.

    Requires use of the PHAC helper's logging configuration and the django_structlog
    RequestMiddleware. The PHAC helper's logging configuration ensures these context
    vars are rendered when logging, and the django_structlog RequestMiddleware handles
    clearing the structlog contextvars between requests.
    """
    # fairly trivial implementation, but that's only a side effect of our logging
    # configuration, and use of django_structlog.middlewares.RequestMiddleware
    # Wrapping in a function so I can document this, and write tests against it

    # First, and more straightforward, binding structlog context vars adds to _all_
    # log output because both our `logging` and `structlog` configs share the
    # `structlog.contextvars.merge_contextvars` processor

    # Second, and more obscurely, these bound context vars only apply to the
    # _current_ request because django_structlog.middlewares.RequestMiddleware clears
    # all context vars at the end of its own response logging handler. Not well documented,
    # but expected to be stable. Clearing context vars either on new request or when finished
    # a response is recommended by structlog's own docs, and is the responsibility of any
    # framework-specific structlog wrapper
    # https://www.structlog.org/en/stable/contextvars.html#context-variables
    # https://github.com/jrobichaud/django-structlog/blob/89fdc7d8adb3cb91848f3b2856e01e5d49649d67/django_structlog/middlewares/request.py#L51

    # Modules that read global state are best deffered to call time rather than module-load
    from django.conf import settings  # pylint: disable=import-outside-toplevel

    # pylint: disable=import-outside-toplevel
    from phac_aspc.django.helpers.logging.configure_logging import (
        is_phac_helper_logging_configuration_being_used,
    )

    if (
        "django_structlog.middlewares.RequestMiddleware"
        not in settings.MIDDLEWARE
    ):
        raise MiddlewareNotUsed(
            "django_structlog.middlewares.RequestMiddleware is required for this utility"
        )

    if not is_phac_helper_logging_configuration_being_used:
        raise ImproperlyConfigured(
            "The PHAC helper's logging configuration is required for this utility"
        )

    structlog.contextvars.bind_contextvars(**context_dict)



# Mostly copied from https://github.com/ulule/django-separatedvaluesfield
# too many modifications to use the original code

from django.core import exceptions, validators
from django.db import models
from django.forms.fields import MultipleChoiceField
from django.utils.encoding import force_str
from django.utils.hashable import make_hashable
from django.utils.text import capfirst


class Creator:
    """
    A placeholder class that provides a way to set the attribute on the model.
    """

    def __init__(self, field):
        self.field = field

    # pylint: disable=redefined-builtin
    def __get__(self, obj, type=None):
        if obj is None:
            return self
        return obj.__dict__[self.field.name]

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = self.field.to_python(value)


class BaseSeparatedValuesField(models.Field):
    def __init__(self, *args, **kwargs):
        self.token = kwargs.pop("token", ",")
        self.cast = kwargs.pop("cast", str)

        if kwargs.get("null", False):
            raise exceptions.ImproperlyConfigured(
                "null is not supported for CommaSeparatedTextField, forcing it to False "
            )

        kwargs["null"] = False

        if not kwargs.get("default", None):
            kwargs["default"] = ""

        super().__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, *args, **kwargs):
        # super() assigns this method, we want to know
        # if the class already had a custom display method before this
        has_display_overide = hasattr(cls, f"get_{self.name}_display")
        super().contribute_to_class(cls, name, *args, **kwargs)

        setattr(cls, self.name, Creator(self))

        if self.choices and not has_display_overide:

            def get_display_value(model_instance):
                choices_dict = dict(make_hashable(self.flatchoices))
                # force_str() to coerce lazy strings.

                def get_formated_value(value):
                    return force_str(
                        choices_dict.get(make_hashable(value), value),
                        strings_only=True,
                    )

                values = getattr(model_instance, self.attname)
                pretty_values = [
                    get_formated_value(value) for value in values if value
                ]

                return ", ".join(pretty_values)

            setattr(cls, f"get_{self.name}_display", get_display_value)

    def validate(self, value, model_instance):
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if self.choices and value:
            choices = []

            for option_key, option_value in self.choices:
                if isinstance(option_value, (list, tuple)):
                    # This is an optgroup, so look inside the group for
                    # options.
                    for optgroup_key, _optgroup_value in option_value:
                        choices.append(optgroup_key)
                else:
                    choices.append(option_key)

            # If we have integers, convert them first to be sure we only compare
            # right types
            choices = [self.cast(choice) for choice in choices]

            for val in value:
                if val and val not in choices:
                    raise exceptions.ValidationError(
                        self.error_messages["invalid_choice"] % val
                    )

        if value is None and not self.null:
            raise exceptions.ValidationError(self.error_messages["null"])

        if not self.blank and value in validators.EMPTY_VALUES:
            raise exceptions.ValidationError(self.error_messages["blank"])

    def to_python(self, value):
        if not value:
            return tuple()

        values = value
        if isinstance(value, str):
            values = value.split(self.token)

        return tuple(self.cast(v) for v in values)

    def get_db_prep_value(self, value, *args, **kwargs):
        if isinstance(value, (list, tuple)):
            return self.token.join([s + "" for s in value])

        if not value:
            return value

        raise ValueError(f"Unexpected value type: {type(value)}, {value}")

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_db_prep_value(value)

    def formfield(self, *args, form_class=MultipleChoiceField, **kwargs):
        defaults = {
            "required": not self.blank,
            "label": capfirst(self.verbose_name),
            "help_text": self.help_text,
        }
        if self.has_default():
            if callable(self.default):
                defaults["initial"] = self.default
                defaults["show_hidden_initial"] = True
            else:
                defaults["initial"] = self.get_default()

        if self.choices:
            defaults["choices"] = self.get_choices(include_blank=False)

            for k in list(kwargs):
                if k not in (
                    "choices",
                    "required",
                    "widget",
                    "label",
                    "initial",
                    "help_text",
                    "error_messages",
                    "show_hidden_initial",
                ):
                    del kwargs[k]
        defaults.update(kwargs)
        return form_class(**defaults)


class CommaSeparatedCharField(BaseSeparatedValuesField, models.CharField):
    pass


class CommaSeparatedTextField(BaseSeparatedValuesField, models.TextField):
    pass



from django.db import transaction
from django.test.client import Client

import pytest

from testapp.models import User


@pytest.fixture(autouse=True)
def enable_db_access_for_all_tests(db):
    """
    without this, tests (including old-style) have to explicitly declare db as a dependency
    https://pytest-django.readthedocs.io/en/latest/faq.html#how-can-i-give-database-access-to-all-my-tests-without-the-django-db-marker
    """
    pass


@pytest.fixture(scope="session")
def globally_scoped_fixture_helper(django_db_setup, django_db_blocker):
    with django_db_blocker.unblock():
        # Wrap in try + atomic block to do non crashing rollback
        # This means we don't have to re-create a test DB each time
        try:
            with transaction.atomic():
                yield
                raise Exception
        except Exception:
            pass


@pytest.fixture
def vanilla_user():
    return User.objects.create(username="vanilla_user")


@pytest.fixture
def vanilla_user_client(vanilla_user):
    client = Client()
    client.force_login(vanilla_user)
    return client



"""
Utilities for basic python stuff, e.g. data/string manipulation
"""

from collections import defaultdict


def group_by(iterable, key: callable):
    """
    like builtin groupby, but doesn't require sorted input
    returns a dict of lists
    """
    groups = defaultdict(list)
    for item in iterable:
        groups[key(item)].append(item)
    return groups


def flatten(iterable):
    """flattens an iterable of iterables into a single list"""
    return [item for sublist in iterable for item in sublist]


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()



from urllib.parse import quote, urlencode, urlparse, urlunparse

from django.templatetags.static import static
from django.urls import reverse
from django.utils.translation import activate, get_language, override

from jinja2 import pass_context
from jinja2.ext import Extension, nodes

import phac_aspc.django.helpers.templatetags as phac_aspc
from phac_aspc.jinja.registry import registry as r

# basic python builtins
r.add_global("getattr", getattr)
r.add_global("hasattr", hasattr)
r.add_global("len", len)
r.add_global("list", list)
r.add_global("print", print)
r.add_global("True", True)
r.add_global("False", False)


# django helpers
r.add_global("static", static)
r.add_global("phac_aspc", phac_aspc)
r.add_global("urlencode", urlencode)
r.add_global("url", reverse)


@r.add_extension
class LanguageExtension(Extension):
    """
    manually override language for a block, e.g. :
        {% language 'fr-ca' %}
        <p>{{ _("Hello") }}</p>
        {% endlanguage %}
    """

    tags = {"language"}

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        # Parse the language code argument
        args = [parser.parse_expression()]
        # Parse everything between the start and end tag:
        body = parser.parse_statements(["name:endlanguage"], drop_needle=True)
        # Call the _switch_language method with the given language code and body
        return nodes.CallBlock(
            self.call_method("_switch_language", args), [], [], body
        ).set_lineno(lineno)

    def _switch_language(self, language_code, caller):
        with override(language_code):
            # Temporarily override the active language and render the body
            output = caller()
        return output


@r.add_global
def convert_url_other_lang(url_str):
    parsed_url = urlparse(url_str)
    path = parsed_url.path
    query = parsed_url.query

    if "fr-ca" in path:
        new_path = path.replace("/fr-ca", "")
    else:
        new_path = "/fr-ca" + path

    new_url = parsed_url._replace(path=new_path)

    if "login" in path and "next" in query:
        if "fr-ca" in path:
            new_query = query.replace("next=/fr-ca", "next=")
        else:
            new_query = query.replace("next=", "next=/fr-ca")
    else:
        new_query = query

    new_url = new_url._replace(query=new_query)

    return urlunparse(new_url)


@r.add_global
@pass_context
def url_to_other_lang(context):
    """
    Provides the URL to the other language:
    For example, if current language is English then it will provide
    the url to the French language.
    """
    request = context["request"]
    full_uri = request.get_full_path()
    return convert_url_other_lang(full_uri)


@r.add_global
def get_lang_code():
    """
    Provides the language code (en-ca or fr-ca) for the current language
    """
    current_lang = get_language()
    return current_lang.lower()


@r.add_global
def get_other_lang_code():
    """
    Provides the language code (en-ca or fr-ca) for the other language (Ex. if current lang
    is en-ca, then the other lang is fr-ca), this is currently used for
    setting the lang tag in the button switch UI
    """
    current_lang = get_language()
    if "en" in current_lang.lower():
        return "fr-ca"
    return "en-ca"


@r.add_global
def get_other_lang():
    """
    Returns the language label ("Français" or "English") for the other language not currently being used.
    """
    current_lang = get_language()
    if "en" in current_lang.lower():
        return "Français"
    return "English"


@r.add_global
@pass_context
def ipython(context):
    import IPython

    IPython.embed()
    return ""


@r.add_global
def cls_str(*args, **kwargs):
    """
    Utility for manipulating many class names
    padds a space to the beginning and end of the string
    """

    args_str = " ".join(str(arg) for arg in args if arg)
    kwargs_str = " ".join(k for k, v in kwargs.items() if v and k)

    s = ""

    if args_str:
        s += " " + args_str

    if kwargs_str:
        s += " " + kwargs_str

    if not s:
        return ""

    return s + " "


@r.add_global
@pass_context
def respects_rule(context, rule, obj=None):
    from phac_aspc.rules import test_rule

    user = context["request"].user
    if not user.is_authenticated:
        return False
    return test_rule(rule, user, obj)



"""Setup URLs for views related to authentication"""

from django.conf.urls.i18n import i18n_patterns
from django.urls import path

from phac_aspc.django.helpers.auth.views import authorize, login
from phac_aspc.django.settings.security_env import get_oauth_env_value

urlpatterns = (
    i18n_patterns(
        path("phac_aspc_helper_login", login, name="phac_aspc_helper_login"),
        path(
            "phac_aspc_helper_authorize", authorize, name="phac_aspc_authorize"
        ),
    )
    if get_oauth_env_value("PROVIDER")
    else []
)



from django.db import connection
from django.forms.models import ModelForm

from testapp.models import TAG_CATEGORIES, Tag


def test_csv_field_on_create():
    tag = Tag.objects.create(
        tag_categories=["fiction", "biography"],
        tag_categories_text=["fiction", "history"],
    )
    assert tag.tag_categories == ("fiction", "biography")
    assert tag.tag_categories_text == ("fiction", "history")
    assert Tag.objects.filter(
        tag_categories="fiction,biography",
        tag_categories_text="fiction,history",
    ).exists()


def test_csv_char_field():
    tag = Tag.objects.create()

    assert tag.tag_categories == tuple()
    tag.tag_categories = [
        "fiction",
        "history",
    ]
    tag.save()

    tag.refresh_from_db()
    assert tag.tag_categories == (
        "fiction",
        "history",
    )
    assert Tag.objects.filter(tag_categories="fiction,history").exists()

    # test forms work too
    class SystemOverviewForm(ModelForm):
        class Meta:
            model = Tag
            fields = ["tag_categories"]

    read_form = SystemOverviewForm(instance=tag)
    assert read_form.initial["tag_categories"] == (
        "fiction",
        "history",
    )
    assert read_form.fields["tag_categories"].choices[0][0] == "fiction"

    bad_form = SystemOverviewForm(
        instance=tag,
        data={
            "tag_categories": [
                "fiction",
                "bad_option",
            ]
        },
    )
    assert not bad_form.is_valid()

    write_form = SystemOverviewForm(
        instance=tag,
        data={
            "tag_categories": [
                "fiction",
                "history",
            ]
        },
    )
    assert write_form.is_valid()
    write_form.save()

    tag.refresh_from_db()
    assert tag.tag_categories == (
        "fiction",
        "history",
    )

    # now check the raw DB value is "" if we set it to []
    tag.tag_categories = []
    tag.save()

    # perform a raw query
    with connection.cursor() as cursor:
        cursor.execute(
            f"SELECT tag_categories FROM testapp_tag WHERE id= {tag.id}"
        )
        row = cursor.fetchone()

    assert row == ("",)


def test_csv_text_field():
    tag = Tag.objects.create()

    assert tag.tag_categories_text == tuple()
    tag.tag_categories_text = [
        "fiction",
        "history",
    ]
    tag.save()

    tag.refresh_from_db()
    assert tag.tag_categories_text == (
        "fiction",
        "history",
    )
    assert Tag.objects.filter(tag_categories_text="fiction,history").exists()

    # test forms work too
    class SystemOverviewForm(ModelForm):
        class Meta:
            model = Tag
            fields = ["tag_categories_text"]

    read_form = SystemOverviewForm(instance=tag)
    assert read_form.initial["tag_categories_text"] == (
        "fiction",
        "history",
    )
    assert read_form.fields["tag_categories_text"].choices[0][0] == "fiction"

    bad_form = SystemOverviewForm(
        instance=tag,
        data={
            "tag_categories_text": [
                "fiction",
                "bad_option",
            ]
        },
    )
    assert not bad_form.is_valid()

    write_form = SystemOverviewForm(
        instance=tag,
        data={
            "tag_categories_text": [
                "fiction",
                "history",
            ]
        },
    )
    assert write_form.is_valid()
    write_form.save()

    tag.refresh_from_db()
    assert tag.tag_categories_text == (
        "fiction",
        "history",
    )

    # now check the raw DB value is "" if we set it to []
    tag.tag_categories_text = tuple()
    tag.save()

    # perform a raw query
    with connection.cursor() as cursor:
        cursor.execute(
            f"SELECT tag_categories_text FROM testapp_tag WHERE id= {tag.id}"
        )
        row = cursor.fetchone()

    assert row == ("",)



from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils import timezone

from phac_aspc.django import fields


class User(AbstractUser):
    pass


class Author(models.Model):
    first_name = fields.CharField(max_length=100)
    last_name = fields.CharField(max_length=100)


TAG_CATEGORIES = [
    ("fiction", "Fiction"),
    ("non-fiction", "Non-Fiction"),
    ("biography", "Biography"),
    ("history", "History"),
]


class Tag(models.Model):
    name = fields.CharField(max_length=250)

    tag_categories = fields.CommaSeparatedCharField(
        max_length=250, blank=True, choices=TAG_CATEGORIES
    )
    tag_categories_text = fields.CommaSeparatedTextField(
        blank=True, choices=TAG_CATEGORIES
    )

    def __str__(self):
        return self.name


class Book(models.Model):
    # changelog_live_name_loader_class = BookNameLoader

    author = fields.ForeignKey(
        Author, related_name="books", on_delete=models.CASCADE
    )
    title = fields.CharField(max_length=250)
    tags = fields.ManyToManyField(Tag)



"""
Django application entrypoint
"""

from django.apps import AppConfig

from phac_aspc.django.helpers.ready import process_ready_hooks


class HelpersConfig(AppConfig):
    """
    Django application config.  As long as this application is listed
    last in INSTALLED_APPS, the `ready` function below will be executed
    when all other applications have been loaded.
    """

    default_auto_field = "django.db.models.BigAutoField"
    name = "phac_aspc.django.helpers"

    def ready(self):
        process_ready_hooks()



"""Security and authentication configuration env var configs and getters"""

from phac_aspc.django.settings.utils.env_utils import (
    PHAC_ENV_PREFIX,
    get_env,
    get_env_value,
)

OAUTH_ENV_PREFIX = f"{PHAC_ENV_PREFIX}OAUTH_"

oauth_env = get_env(
    prefix=OAUTH_ENV_PREFIX,
    PROVIDER=(str, ""),
    APP_CLIENT_ID=(str, ""),
    APP_CLIENT_SECRET=(str, ""),
    MICROSOFT_TENANT=(str, "common"),
    REDIRECT_ON_LOGIN=(str, ""),
    USE_BACKEND=(
        str,
        "phac_aspc.django.helpers.auth.backend.PhacAspcOAuthBackend",
    ),
)


def get_oauth_env_value(key):
    return get_env_value(oauth_env, key, prefix=OAUTH_ENV_PREFIX)


security_env_config = {
    # ----- AC-11 - Session controls
    # Sessions expire in 20 minutes
    "SESSION_COOKIE_AGE": (int, 1200),
    # Use HTTPS for session cookie
    "SESSION_COOKIE_SECURE": (bool, True),
    # Sessions close when browser is closed
    "SESSION_EXPIRE_AT_BROWSER_CLOSE": (bool, True),
    # Every requests extends the session (This is required for the WET session
    # plugin to function properly.)
    "SESSION_SAVE_EVERY_REQUEST": (bool, True),
}



import io
import random
from unittest.mock import Mock, patch

from django.test import RequestFactory

import pytest
from openpyxl import Workbook
from openpyxl.styles import (
    Alignment,
    Border,
    Font,
    NamedStyle,
    PatternFill,
    Side,
)

from phac_aspc.django.excel import (
    AbstractCsvExportView,
    AbstractExportView,
    AbstractSheetWriter,
    CustomColumn,
    ManyToManyColumn,
    ModelColumn,
    ModelToCsvWriter,
    ModelToSheetWriter,
    WriterConfigException,
)
from testapp.model_factories import AuthorFactory, BookFactory, TagFactory
from testapp.models import Author, Book


def make_request():
    req_factory = RequestFactory()
    return req_factory.get("/fake-url")


def instantiate_view(ViewClass):
    request = make_request()
    view_instance = ViewClass()
    view_instance.setup(request)
    return view_instance


def get_response(ViewClass):
    request = make_request()
    view_func = ViewClass.as_view()
    return view_func(request)


def create_data():
    all_tags = [TagFactory() for _ in range(6)]

    for _ in range(30):
        book = BookFactory()
        tags = random.sample(all_tags, random.randint(0, 4))
        book.tags.set(tags)


def test_model_to_sheet_writer(django_assert_max_num_queries):
    create_data()

    columns = [
        ModelColumn(Book, "title"),
        CustomColumn(
            "Author", lambda x: f"{x.author.first_name} {x.author.last_name}"
        ),
        ManyToManyColumn(Book, "tags"),
    ]

    class BookSheetWriter(ModelToSheetWriter):

        def get_column_configs(self):
            return columns

    with django_assert_max_num_queries(4):
        wb = Workbook(write_only=True)
        writer = BookSheetWriter(
            workbook=wb,
            queryset=Book.objects.all().prefetch_related("author", "tags"),
            sheet_name="books",
        )
        writer.write()


def test_abstract_view():
    create_data()

    qs = Book.objects.all().prefetch_related("author", "tags")

    class BookExportView(AbstractExportView):
        sheetwriter_class = ModelToSheetWriter
        queryset = qs

    view_func = BookExportView.as_view()

    view_instance = instantiate_view(BookExportView)
    assert view_instance.get_sheetwriter_class() == ModelToSheetWriter
    assert view_instance._get_iterable() == qs

    response = get_response(BookExportView)
    assert response.status_code == 200

    # now test it works if queryset is set to None
    # e.g. that it uses the writer's queryset

    class BookSheetQsWriterWithQs(ModelToSheetWriter):
        queryset = qs

    class BookExportViewWithoutQs(AbstractExportView):
        sheetwriter_class = BookSheetQsWriterWithQs

    view_instance = instantiate_view(BookExportViewWithoutQs)
    assert view_instance.get_sheetwriter_class() == BookSheetQsWriterWithQs
    assert view_instance._get_iterable() is None

    response = get_response(BookExportViewWithoutQs)
    assert response.status_code == 200


def test_abstract_view_iterable_behaviour():

    class BookExportView(AbstractCsvExportView):
        def get_iterator(self):
            return range(10)

    view_instance = instantiate_view(BookExportView)
    assert view_instance._get_iterable() == range(10)

    class BookExportView2(AbstractCsvExportView):
        iterator = range(10)

    view_instance = instantiate_view(BookExportView2)
    assert view_instance._get_iterable() == range(10)

    qs = Book.objects.all()

    class BookExportView3(AbstractCsvExportView):
        def get_queryset(self):
            return qs

    view_instance = instantiate_view(BookExportView3)
    assert view_instance._get_iterable() is qs

    class BookExportView4(AbstractCsvExportView):
        queryset = qs

    view_instance = instantiate_view(BookExportView4)
    assert view_instance._get_iterable() is qs


def test_abstract_view_with_non_qs_writer():
    class BookSheetWriter(AbstractSheetWriter):
        sheet_name = "Books"

        def get_column_configs(self):
            return [
                ModelColumn(Book, "title"),
                CustomColumn(
                    "Author",
                    lambda x: f"{x.author.first_name} {x.author.last_name}",
                ),
                ManyToManyColumn(Book, "tags"),
            ]

    class BookExportView(AbstractExportView):
        def get_sheetwriter_class(self):
            return BookSheetWriter

        def get_iterator(self):
            return list(Book.objects.all().prefetch_related("author", "tags"))

    view_inst = instantiate_view(BookExportView)
    assert view_inst.get_sheetwriter_class() == BookSheetWriter
    assert view_inst._get_iterable() == list(
        Book.objects.all().prefetch_related("author", "tags")
    )

    response = get_response(BookExportView)
    assert response.status_code == 200

    # also test with null-iterator, e.g. so it calls the writer's get_iterator
    class BookSheetWriterWithIterator(BookSheetWriter):
        iterator = list(Book.objects.all().prefetch_related("author", "tags"))

    class BookExportViewWithoutIterator(AbstractExportView):
        sheetwriter_class = BookSheetWriterWithIterator

    view_inst = instantiate_view(BookExportViewWithoutIterator)
    assert view_inst._get_iterable() is None

    response = get_response(BookExportViewWithoutIterator)
    assert response.status_code == 200


def test_csv_writer():
    class BookCsvWriter(ModelToCsvWriter):
        model = Book

        def get_column_configs(self):
            return [
                ModelColumn(Book, "title"),
                CustomColumn(
                    "Author",
                    lambda x: f"{x.author.first_name} {x.author.last_name}",
                ),
                ManyToManyColumn(Book, "tags", delimiter="|"),
            ]

    # test view response
    t1 = TagFactory(name="t1")
    t2 = TagFactory(name="t2")
    a1 = AuthorFactory(first_name="bôb", last_name="l'ébob")
    b1 = BookFactory(title="b1", author=a1)
    b1.tags.set([t1, t2])
    b2 = BookFactory(title="b2 çûèêëcks", author=a1)

    # test serialization
    # TODO: figure out why testing response content directly fails
    # for some reason it escapes utf8
    file = io.StringIO()
    writer = BookCsvWriter(
        file, queryset=Book.objects.filter(id__in=[b1.id, b2.id])
    )
    writer.write()
    as_str = file.getvalue()
    assert (
        as_str
        == "title,Author,tags\r\nb1,bôb l'ébob,t1|t2\r\nb2 çûèêëcks,bôb l'ébob,\r\n"
    )


def test_abstract_csv_view():
    constructor_spy = Mock()
    write_spy = Mock()

    class WriterClassMock:
        def __init__(self, *args, **kwargs):
            constructor_spy(*args, **kwargs)

        def write(self):
            write_spy()

    qs = Book.objects.all().prefetch_related("author", "tags")

    class CsvExportView(AbstractCsvExportView):
        writer_class = WriterClassMock
        queryset = qs

    response = get_response(CsvExportView)
    assert response.status_code == 200
    constructor_spy.assert_called_with(iterator=qs, buffer=response)
    write_spy.assert_called_once()


def test_abstract_csv_view_with_qs_writer():
    qs = Book.objects.all().prefetch_related("author", "tags")

    class CsvExportView(AbstractCsvExportView):
        writer_class = ModelToCsvWriter
        queryset = qs

    response = get_response(CsvExportView)
    assert response.status_code == 200

    # now test again, but without queryset on the view class

    class BookCsvWriter(ModelToCsvWriter):
        queryset = qs

    class CsvExportViewWithoutQs(AbstractCsvExportView):
        writer_class = BookCsvWriter

    response = get_response(CsvExportViewWithoutQs)
    assert response.status_code == 200


def test_abstract_excel_view():
    constructor_spy = Mock()
    write_spy = Mock()

    class WriterClassMock:
        def __init__(self, *args, **kwargs):
            constructor_spy(*args, **kwargs)

        def write(self):
            write_spy()

    wbInstanceMock = Mock()

    qs = Book.objects.all().prefetch_related("author", "tags")

    class CsvExportView(AbstractExportView):
        sheetwriter_class = WriterClassMock
        queryset = qs

    view_func = CsvExportView.as_view()
    req_factory = RequestFactory()
    request = req_factory.get("/fake-url")

    with patch("openpyxl.Workbook", return_value=wbInstanceMock):
        response = view_func(request)

    assert response.status_code == 200

    constructor_spy.assert_called_with(iterator=qs, workbook=wbInstanceMock)
    write_spy.assert_called_once()
    wbInstanceMock.save.assert_called_once()
    wbInstanceMock.save.assert_called_with(response)


def test_various_writer_apis():
    wb = Workbook(write_only=True)
    qs = Book.objects.all().prefetch_related("author", "tags")

    class BookWriter(ModelToSheetWriter):
        model = Book

    with pytest.raises(WriterConfigException):
        # iterator kwarg should fail at __init__
        BookWriter(workbook=wb, iterator=qs)

    # try with qs kwarg
    BookWriter(workbook=wb, queryset=qs).write()

    with pytest.raises(TypeError):
        # should not take any positional args
        BookWriter(qs, workbook=wb)

    class BookWriterWithGetQuerysetMethod(BookWriter):
        def get_queryset(self):
            return qs

    BookWriterWithGetQuerysetMethod(workbook=wb).write()

    class BookWriterWithQuerysetAttribute(BookWriter):
        queryset = qs

    BookWriterWithQuerysetAttribute(workbook=wb).write()

    class BookWriterWithGetIteratorMethod(BookWriter):
        def get_iterator(self):
            return qs

    with pytest.raises(WriterConfigException):
        BookWriterWithGetIteratorMethod(
            workbook=wb, sheet_name="books"
        ).write()

    class BookWriterWithIteratorAttribute(BookWriter):
        """
        model-writers cannot define the iterator directly
        """

        iterator = qs

    with pytest.raises(WriterConfigException):
        BookWriterWithIteratorAttribute(workbook=wb).write()

    class BookWriterWithGetIteratorMethodAndSheetName(BookWriter):
        queryset = qs
        model = Book
        sheet_name = "books"

    BookWriterWithGetIteratorMethodAndSheetName(workbook=wb).write()


def test_writer_attr_precedence():
    # sheet_name and queryset kwargs take precedence over class attributes
    wb = Workbook(write_only=True)
    book_qs = Book.objects.all().prefetch_related("author", "tags")

    author_qs = Author.objects.all()

    class BookWriter(ModelToSheetWriter):
        queryset = book_qs
        sheet_name = "books"

    author_writer = BookWriter(
        workbook=wb, queryset=author_qs, sheet_name="authors"
    )
    assert author_writer.queryset == author_qs
    assert author_writer.sheet_name == "authors"


def test_writer_get_columns_api():
    create_data()
    wb = Workbook(write_only=True)
    qs = Book.objects.all().prefetch_related("author", "tags")

    column_defs = [
        ModelColumn(Book, "title"),
        CustomColumn(
            "Author", lambda x: f"{x.author.first_name} {x.author.last_name}"
        ),
        ManyToManyColumn(Book, "tags"),
    ]
    column_defs2 = column_defs[:1]

    class BookWriterWithColumnConfigsMethod(ModelToSheetWriter):
        def get_column_configs(self):
            return column_defs

    writer1 = BookWriterWithColumnConfigsMethod(workbook=wb, queryset=qs)
    assert writer1.get_column_configs() is column_defs
    # if you override get_column_configs, get_columns is still the old method
    assert not writer1.get_columns() is column_defs

    class BookWriterWithColumnsMethod(ModelToSheetWriter):
        def get_columns(self):
            return column_defs

    writer2 = BookWriterWithColumnsMethod(workbook=wb, queryset=qs)
    assert writer2.get_column_configs() is column_defs
    assert writer2.get_columns() is column_defs

    class BookWriterWithColumnClassAttr(ModelToSheetWriter):
        columns = column_defs

    writer3 = BookWriterWithColumnClassAttr(workbook=wb, queryset=qs)
    assert writer3.get_column_configs() is column_defs
    assert writer3.get_columns() is column_defs

    # the constructor kwarg should take precedence over the class attr
    writer4 = BookWriterWithColumnClassAttr(
        workbook=wb, queryset=qs, columns=column_defs2
    )
    assert writer4.get_column_configs() is column_defs2
    assert writer4.get_columns() is column_defs2

    class BookWriterWithNoCols(ModelToSheetWriter):
        pass

    writer5 = BookWriterWithNoCols(workbook=wb, queryset=qs)
    # model sheet writer has its own default get_columns fallback
    assert writer5.get_column_configs() is not None
    assert writer5.get_columns() is not None

    # non-model writers should raise error if no column configs are provided
    class NonModelWriter(AbstractSheetWriter):
        pass

    with pytest.raises(NotImplementedError):
        NonModelWriter(workbook=wb, sheet_name="abc").get_column_configs()


def test_sheetwriter_with_column_styles():
    create_data()
    wb = Workbook(write_only=True)
    qs = Book.objects.all().prefetch_related("author", "tags")

    bold_style = NamedStyle(name="bold", font=Font(bold=True))
    # bold_style.font = Font(bold=True)

    class BookWriterWithStyles(ModelToSheetWriter):
        columns = [
            ModelColumn(Book, "title", style=bold_style, column_width=50),
            CustomColumn(
                "Author",
                lambda x: f"{x.author.first_name} {x.author.last_name}",
            ),
            ManyToManyColumn(Book, "tags"),
        ]

    writer = BookWriterWithStyles(workbook=wb, queryset=qs)
    writer.write()


def test_sheetwriter_with_header_style():

    create_data()
    wb = Workbook(write_only=True)
    qs = Book.objects.all().prefetch_related("author", "tags")

    border_side = Side(style="thin", color="000000")
    style = NamedStyle(
        name="header",
        font=Font(bold=True),
        alignment=Alignment(horizontal="center"),
        fill=PatternFill("solid", fgColor="00C0C0C0"),
        border=Border(
            left=border_side,
            top=border_side,
            right=border_side,
            bottom=border_side,
        ),
    )
    # header_style.font = Font(bold=True)
    # header_style.alignment = Alignment(horizontal="center")
    # header_style.fill = PatternFill("solid", fgColor="00C0C0C0")
    # header_style.border = Border(left=bd, top=bd, right=bd, bottom=bd)

    class BookWriterWithHeaderStyle(ModelToSheetWriter):
        header_style = style

        columns = [
            ModelColumn(Book, "title"),
            CustomColumn(
                "Author",
                lambda x: f"{x.author.first_name} {x.author.last_name}",
            ),
            ManyToManyColumn(Book, "tags"),
        ]

    BookWriterWithHeaderStyle(workbook=wb, queryset=qs).write()



"""This file contains a utility for configuring PHAC helpers settings when
in a testing environment
"""


def configure_settings_for_tests():
    """
    Modify django settings so they are suitable to run unit tests.
      - Disables the Axes authentcation backend, to allow the test client
        to function.  You could instead modify the request as described here:
        https://django-axes.readthedocs.io/en/latest/3_usage.html#authenticating-users
    """

    # Modules that read global state are best deffered to call time rather than module-load
    from django.conf import settings  # pylint: disable=import-outside-toplevel

    # settings.AXES_ENABLED = False

    settings.AUTHENTICATION_BACKENDS = [
        x
        for x in settings.AUTHENTICATION_BACKENDS
        if x != "axes.backends.AxesStandaloneBackend"
    ]
